#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

"""Support for making and reading "tree shapes".

A "tree shape" is a dict mapping names of files and directories to attributes,
where "attributes" are dicts that contain keys like ``CONTENT`` or
``PERMISSIONS`` that describe the file or directory.

In a tree shape, if a key ends with a '/', then it's a directory.  Otherwise,
it's a file.
"""

import errno
import os


# This is what the standard library uses for the default mode passed to
# os.makedirs.
DEFAULT_MODE = 0777

CONTENT = 'content'
PERMISSIONS = 'permissions'


class InvalidDirectory(ValueError):
    """Raised when we are given a confusing directory description."""

    def __init__(self, entry):
        super(InvalidDirectory, self).__init__(
            "Directories must end with '/' and have no content, got %r"
            % (entry,))


def validate_entry(name, attributes, original=None):
    is_dir = (name[-1] == '/')
    if is_dir:
        if attributes.get(CONTENT, None) is not None:
            if not original:
                original = (name, attributes)
            raise InvalidDirectory(original)


def make_tree(base_directory, shape, mode=DEFAULT_MODE):
    """Make a tree of files and directories underneath ``base_directory``.

    :param shape: A dict of descriptions of files and directories to make.
        The name of the file or directory to create is the key.  If you want
        to create a directory, make sure the name ends with '/'.

        The value is a dict.  Recognized keys are ``PERMISSIONS``, which if
        provided will be used to set the file system permissions for the
        created directory, and ``CONTENT``, which if provided for a file
        will be the contents of that file.

    :raise InvalidDirectory: if a directory has a ``CONTENT`` key.
    """
    for name, attributes in sorted(shape.items()):
        if not attributes:
            attributes = {}
        path = os.path.join(base_directory, name)
        validate_entry(name, attributes)
        if name[-1] == '/':
            os.makedirs(path, mode)
        else:
            base_dir = os.path.dirname(path)
            try:
                os.makedirs(base_dir, mode)
            except OSError, e:
                if e.errno != errno.EEXIST:
                    raise
            contents = attributes.get(CONTENT, "The file '%s'." % (name,))
            f = open(path, 'w')
            f.write(contents)
            f.close()
        permissions = attributes.get(PERMISSIONS)
        if permissions:
            os.chmod(path, permissions)


def _get_permissions(path):
    """Get the permissions of ``path``.

    Much like os.stat(path).st_mode, except we only preserve the last four
    octal digits.  Thus, it returns things like 0755 rather than 0100755.
    """
    return int(oct(os.stat(path).st_mode)[-4:], 8)


def _get_info(dirpath, name, directory):
    path = os.path.join(dirpath, name)
    name = os.path.relpath(path, directory)
    permissions = _get_permissions(path)
    return path, name, permissions


def load_tree(directory):
    """Return a tree shape corresponding to the files beneath ``directory``.

    :param directory: A directory on disk.
    :raises ValueError: If ``directory`` points to a non-directory.
    :return: a tree shape of all files and directories beneath ``directory``.
    """
    if not os.path.isdir(directory):
        raise ValueError(
            "Can only load tree for directories, got %r" % (directory,))
    shape = {}
    for dirpath, dirnames, filenames in os.walk(directory):
        for name in dirnames:
            path, name, permissions = _get_info(dirpath, name, directory)
            shape[name + '/'] = {PERMISSIONS: permissions}
        for name in filenames:
            path, name, permissions = _get_info(dirpath, name, directory)
            content = open(path, 'rb').read()
            shape[name] = {CONTENT: content, PERMISSIONS: permissions}
    return shape



#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.


from ._treeshape import (
    CONTENT,
    PERMISSIONS,
    validate_entry,
    )


def _is_permission(possible_permissions):
    """Is ``possible_permissions`` a permission?"""
    return isinstance(possible_permissions, int)


def _guess_content_or_permissions(content_or_permissions):
    """
    Given a thing that could be either content or permissions, guess which it
    is.

    Used when we are given a 2-tuple to figure out what the second element of
    the tuple actually is.
    """
    if content_or_permissions is None:
        return {}
    elif _is_permission(content_or_permissions):
        return {PERMISSIONS: content_or_permissions}
    else:
        return {CONTENT: content_or_permissions}


def _normalize_entry(entry):
    if isinstance(entry, basestring):
        return (entry, {})

    if len(entry) == 1:
        return (entry[0], {})

    if len(entry) == 2:
        name, content_or_permissions = entry
        attributes = _guess_content_or_permissions(content_or_permissions)
        return name, attributes

    if len(entry) == 3:
        name, content, permissions = entry
        attributes = {}
        if content:
            attributes[CONTENT] = content
        if permissions:
            attributes[PERMISSIONS] = permissions
        return name, attributes

    raise ValueError(
        "Invalid file or directory description: %r" % (entry,))


def _from_rough_entry(entry):
    """Normalize a file shape entry.

    'Normal' entries are ("file": {CONTENT: "content"}), ("directory/",
    {PERMISSIONS: 0755}) and so forth.  They have a name and attributes.

    Standalone strings get turned into normal entries with no attributes.
    Singleton tuples are treated the same.

    If something looks like a directory and has content, we raise an error, as
    we don't know whether the developer really intends a file or really
    intends a directory.

    :raise ValueError: If we're given a tuple that's too long, or that we
        can't understand.
    :raise InvalidDirectory: If we're given a directory description with file
        content.  What's that even supposed to mean?
    :return: A 2-tuple of name and attributes. Attributes might contain
        CONTENT or PERMISSIONS.
    """
    name, attributes = _normalize_entry(entry)
    validate_entry(name, attributes, entry)
    return name, attributes


class DuplicateName(Exception):

    def __init__(self, first, second):
        super(DuplicateName, self).__init__(
            "Duplicate name given.  Did you mean %r or did you mean %r?"
            % (first, second))
        self.first = first
        self.second = second


def from_rough_spec(list_shape):
    """Create a tree shape from a list of rough specifications.

    Rather than fully and explicitly specifying a tree shape,
    ``from_rough_spec`` allows you to say things like::

      dwim(['filename', ('directory/', 0755), ('data', 'content goes here')])

    And have that turned into a proper tree shape::

      {'filename': {},
       'directory/': {PERMISSIONS: 0755},
       'data': {CONTENT: 'content goes here'}}

    Use this when you wish to be succinct.

    Consider this a "Do What I Mean" (DWIM) function, with all the pros and
    cons that entails.

    :raises DuplicateName: If the same name is given more than once.
    :raises ValueError: If any of the specifications are too long, or
        we simply cannot understand.
    :raises InvalidDirectory: If a specification looks like it's for a
        directory, but has content.  What's that even supposed to mean?
    :return: A dict mapping names to attributes.
    """
    shape = {}
    for entry in list_shape:
        name, attributes = _from_rough_entry(entry)
        if name in shape:
            raise DuplicateName((name, shape[name]), entry)
        shape[name] = attributes
    return shape



#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import stat

from testtools import TestCase
from testtools.matchers import (
    Contains,
    MatchesPredicate,
    )

from treeshape import (
    CONTENT,
    FileTree,
    from_rough_spec,
    HasFileTree,
    MatchesFileTree,
    PERMISSIONS,
    )

from treeshape._matchers import (
    add_implicit_directories,
    iterate_parents,
    )


class TestIterateParents(TestCase):

    def test_single_segment(self):
        self.assertEqual([], list(iterate_parents('foo')))
        self.assertEqual([], list(iterate_parents('foo/')))

    def test_multiple_segments(self):
        self.assertEqual(
            [('foo/bar/'), ('foo/')],
            list(iterate_parents('foo/bar/baz.txt')))
        self.assertEqual(
            [('foo/bar/'), ('foo/')],
            list(iterate_parents('foo/bar/baz/')))


class TestAddImplicitDirectories(TestCase):

    def test_empty(self):
        self.assertEqual({}, add_implicit_directories({}))

    def test_no_implicit_directories(self):
        shape = from_rough_spec([
            'foo',
            'bar/',
            'bar/baz.txt',
            ])
        self.assertEqual(shape, add_implicit_directories(shape))

    def test_implicit_directories(self):
        shape = from_rough_spec([
            'foo',
            'bar/baz.txt',
            'bar/qux/',
            ])
        self.assertEqual(
            {'bar/': {},
             'bar/baz.txt': {},
             'bar/qux/': {},
             'foo': {},
            }, add_implicit_directories(shape))


class TestMatchesFileTree(TestCase):

    def test_empty(self):
        path = self.useFixture(FileTree({})).path
        self.assertThat(path, MatchesFileTree({}))

    def test_missing_file(self):
        path = self.useFixture(FileTree({})).path
        mismatch = MatchesFileTree({'foo': {}}).match(path)
        self.assertIsNot(None, mismatch)

    def test_unexpected_file(self):
        path = self.useFixture(FileTree({'foo': {}, 'bar': {}})).path
        mismatch = MatchesFileTree({'bar': {}}).match(path)
        self.assertIsNot(None, mismatch)

    def test_files_and_directories(self):
        shape = {
            'foo': {CONTENT: 'contents', PERMISSIONS: 0755},
            'bar/': {PERMISSIONS: 0700},
            }
        path = self.useFixture(FileTree(shape)).path
        self.assertThat(path, MatchesFileTree(shape))

    def test_with_matchers(self):
        shape = {
            'foo': {CONTENT: 'contents', PERMISSIONS: 0755},
            'bar/': {PERMISSIONS: 0400},
            }
        path = self.useFixture(FileTree(shape)).path
        is_user_readable = MatchesPredicate(
            lambda p: p & stat.S_IRUSR, "User can't read: %s")
        self.assertThat(
            path, MatchesFileTree(
                {'foo': {CONTENT: Contains('on')},
                 'bar/': {PERMISSIONS: is_user_readable}}))

    def test_unspecified_permissions(self):
        shape = {
            'foo': {CONTENT: 'contents'},
            'bar/': {},
            }
        path = self.useFixture(FileTree(shape)).path
        self.assertThat(path, MatchesFileTree(shape))

    def test_unspecified_contents(self):
        shape = {'foo': {}, 'bar/': {}}
        path = self.useFixture(FileTree(shape)).path
        self.assertThat(path, MatchesFileTree(shape))

    def test_implicit_directories(self):
        shape = {'foo/bar/': {}, 'foo/bar.txt': {}}
        path = self.useFixture(FileTree(shape)).path
        self.assertThat(path, MatchesFileTree(shape))

    def test_with_dwim(self):
        shape = {
            'foo': {CONTENT: 'contents', PERMISSIONS: 0755},
            'bar/': {PERMISSIONS: 0700},
            }
        path = self.useFixture(FileTree(shape)).path
        expected = [
            ('foo', Contains('on')),
            'bar/',
            ]
        self.assertThat(path, MatchesFileTree(from_rough_spec(expected)))


class TestHasFileTree(TestCase):

    def test_unexpected_file(self):
        path = self.useFixture(FileTree({'foo': {}, 'bar': {}})).path
        self.assertThat(path, HasFileTree({'bar': {}}))



#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import os

from fixtures import (
    Fixture,
    TempDir,
    )

from ._treeshape import (
    DEFAULT_MODE,
    make_tree,
    )


class FileTree(Fixture):
    """A structure of files and directories on disk."""

    def __init__(self, shape, mode=DEFAULT_MODE):
        """Create a ``FileTree``.

        :param shape: A list of descriptions of files and directories to make.
            Generally directories are described as ``"directory/"`` and
            files are described as ``("filename", contents)``.  Filenames can
            also be specified without contents, in which case we'll make
            something up.

            Directories can also be specified as ``(directory, None)`` or
            ``(directory,)``.
        """
        super(FileTree, self).__init__()
        self.shape = shape
        self._mode = mode

    def setUp(self):
        super(FileTree, self).setUp()
        tempdir = self.useFixture(TempDir())
        self.path = tempdir.path
        make_tree(self.path, self.shape, self._mode)

    def join(self, *segments):
        """Return the absolute path to the given segments under this tree."""
        return os.path.abspath(os.path.join(self.path, *segments))



#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import os

from testtools import TestCase
from testtools.matchers import (
    DirContains,
    DirExists,
    FileContains,
    HasPermissions,
    Mismatch,
    Not,
    )

from treeshape import FileTree


class HasNoAttribute(object):
    """For asserting that an object does not have a particular attribute."""

    def __init__(self, attr_name):
        self._attr_name = attr_name

    def __str__(self):
        return 'HasNoAttribute(%s)' % (self._attr_name,)

    def match(self, obj):
        sentinel = object()
        value = getattr(obj, self._attr_name, sentinel)
        if value is not sentinel:
            return Mismatch(
                "%s is an attribute of %r: %r" % (
                    self._attr_name, obj, value))


class TestFileTree(TestCase):

    def test_no_path_at_start(self):
        # FileTree fixture doesn't create a path at the beginning.
        fixture = FileTree({})
        self.assertThat(fixture, HasNoAttribute('path'))

    def test_creates_directory(self):
        # It creates a temporary directory once set up.  That directory is
        # removed at cleanup.
        fixture = FileTree({})
        fixture.setUp()
        try:
            self.assertThat(fixture.path, DirExists())
        finally:
            fixture.cleanUp()
            self.assertThat(fixture.path, Not(DirExists()))

    def test_out_of_order(self):
        # If a file or a subdirectory is listed before its parent directory,
        # that doesn't matter.  We'll create the directory first.
        fixture = FileTree({'a/b/': {}, 'a/': {}})
        with fixture:
            path = fixture.path
            self.assertThat(path, DirContains(['a']))
            self.assertThat(os.path.join(path, 'a'), DirContains(['b']))
            self.assertThat(os.path.join(path, 'a', 'b'), DirExists())

    def test_not_even_creating_parents(self):
        fixture = FileTree({'a/b/foo.txt': {}, 'c/d/e/': {}})
        with fixture:
            path = fixture.path
            self.assertThat(
                os.path.join(path, 'a', 'b', 'foo.txt'),
                FileContains("The file 'a/b/foo.txt'."))
            self.assertThat(os.path.join(path, 'c', 'd', 'e'), DirExists())

    def test_mode(self):
        fixture = FileTree({'a/b/foo.txt': {}, 'c/d/e/': {}}, mode=0700)
        with fixture:
            path = fixture.path
            self.assertThat(
                os.path.join(path, 'a', 'b'),
                HasPermissions('0700'))

    def test_join(self):
        fixture = FileTree({})
        with fixture:
            path = fixture.join('foo', 'bar', 'baz')
            self.assertEqual(
                os.path.abspath(os.path.join(fixture.path, 'foo', 'bar', 'baz')),
                path)



#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

from testtools import TestCase
from testtools.matchers import Is

from treeshape import (
    CONTENT,
    DuplicateName,
    from_rough_spec,
    InvalidDirectory,
    PERMISSIONS,
    )
from treeshape._dwim import (
    _from_rough_entry,
    )


class TestNormalizeEntry(TestCase):

    def test_file_as_tuple(self):
        # A tuple of filenames and contents is not quite normalized. It still
        # needs 'None' added to the end, to represent default permissions.
        entry = _from_rough_entry(('foo', 'foo contents'))
        self.assertEqual(('foo', {CONTENT: 'foo contents'}), entry)

    def test_file_with_permissions_as_tuple(self):
        # A tuple of filename, contents and permissions is already normalized.
        entry = _from_rough_entry(('foo', 'foo contents', 0755))
        self.assertEqual(
            ('foo', {CONTENT: 'foo contents', PERMISSIONS: 0755}), entry)

    def test_directories_as_tuples(self):
        # A tuple of directory name and None is not quite normalized. It still
        # needs 'None' added to the end, to represent default permissions.
        directory = _from_rough_entry(('foo/', None))
        self.assertEqual(('foo/', {}), directory)

    def test_directories_with_permissions_as_tuples(self):
        # A tuple of directory name, None and None is already normalized.
        directory = _from_rough_entry(('foo/', None, None))
        self.assertEqual(('foo/', {}), directory)

    def test_directories_as_singletons(self):
        # A singleton tuple of directory name is normalized to a 2-tuple of
        # the directory name and None.
        directory = _from_rough_entry(('foo/',))
        self.assertEqual(('foo/', {}), directory)

    def test_directories_as_strings(self):
        # If directories are just given as strings, then they are normalized
        # to tuples of directory names and None.
        directory = _from_rough_entry('foo/')
        self.assertEqual(('foo/', {}), directory)

    def test_directories_with_content(self):
        # If we're given a directory with content, we raise an error, since
        # it's ambiguous and we don't want to guess.
        bad_entry = ('dir/', "stuff")
        e = self.assertRaises(InvalidDirectory, _from_rough_entry, bad_entry)
        self.assertEqual(
            "Directories must end with '/' and have no content, got %r"
            % (bad_entry,), str(e))

    def test_directories_with_permissions(self):
        entry = ('dir/', 0755)
        self.assertEqual(
            ('dir/', {PERMISSIONS: 0755}), _from_rough_entry(entry))

    def test_directories_with_content_and_empty_permissions(self):
        # If we're given a directory with content, we raise an error, since
        # it's ambiguous and we don't want to guess.
        bad_entry = ('dir/', "stuff", None)
        e = self.assertRaises(InvalidDirectory, _from_rough_entry, bad_entry)
        self.assertEqual(
            "Directories must end with '/' and have no content, got %r"
            % (bad_entry,), str(e))

    def test_directories_with_content_and_permissions(self):
        # If we're given a directory with content, we raise an error, since
        # it's ambiguous and we don't want to guess.
        bad_entry = ('dir/', "stuff", 0755)
        e = self.assertRaises(InvalidDirectory, _from_rough_entry, bad_entry)
        self.assertEqual(
            "Directories must end with '/' and have no content, got %r"
            % (bad_entry,),
            str(e))

    def test_filenames_as_strings(self):
        # If file names are just given as strings, then they are normalized to
        # tuples of filenames and made-up contents.
        entry = _from_rough_entry('foo')
        self.assertEqual(('foo', {}), entry)

    def test_filenames_as_singletons(self):
        # A singleton tuple of a filename is normalized to a 2-tuple of
        # the file name and made-up contents.
        entry = _from_rough_entry(('foo',))
        self.assertEqual(('foo', {}), entry)

    def test_filenames_without_content(self):
        # If we're given a filename without content, make something up.
        entry = ('foo', None)
        self.assertEqual(('foo', {}), _from_rough_entry(entry))

    def test_filenames_without_content_with_empty_permissions(self):
        # If we're given a filename without content, we raise an error, since
        # it's ambiguous and we don't want to guess.
        entry = ('filename', None, None)
        self.assertEqual(('filename', {}), _from_rough_entry(entry))

    def test_filenames_with_just_permission(self):
        # If we're given a filename without content, we raise an error, since
        # it's ambiguous and we don't want to guess.
        entry = ('filename', 0755)
        self.assertEqual(
            ('filename', {PERMISSIONS: 0755}), _from_rough_entry(entry))

    def test_filenames_without_content_with_permissions(self):
        # If we're given a filename without content, and we are also given
        # permissions, well then we're honour-bound to create the file.
        entry = ('filename', None, 0755)
        self.assertEqual(
            ('filename', {PERMISSIONS: 0755}), _from_rough_entry(entry))

    def test_too_long_tuple(self):
        bad_entry = ('foo', 'bar', 'baz', 'qux')
        e = self.assertRaises(ValueError, _from_rough_entry, bad_entry)
        self.assertEqual(
            "Invalid file or directory description: %r" % (bad_entry,),
            str(e))

    def test_passes_matchers_through(self):
        marker = object()
        entry = ('filename', Is(marker))
        name, attributes = _from_rough_entry(entry)
        self.assertEqual('filename', name)
        self.assertEqual([CONTENT], attributes.keys())
        self.assertThat(marker, attributes[CONTENT])


class TestFromRoughSpec(TestCase):

    def test_empty(self):
        self.assertEqual({}, from_rough_spec([]))

    def test_dict_of_normalized_entries(self):
        entries = ['filename', ('dir/', 0755)]
        normalized_entries = [('filename', {}), ('dir/', {PERMISSIONS: 0755})]
        self.assertEqual(dict(normalized_entries), from_rough_spec(entries))

    def test_duplicate_name(self):
        entries = ['a', 'a']
        self.assertRaises(DuplicateName, from_rough_spec, entries)



#!/usr/bin/env python

from setuptools import setup, find_packages
import os.path

description = file(os.path.join(os.path.dirname(__file__), 'README'), 'rb').read()

setup(
    name="treeshape",
    version="0.2.1",
    description="Quickly make files and directory structures.",
    long_description=description,
    author="Jonathan Lange",
    author_email="jml@mumak.net",
    install_requires=[
        "fixtures",
        "testtools",
        ],
    zip_safe=False,
    packages=find_packages('.'),
    classifiers = [
        'Development Status :: 6 - Mature',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        ],
)



#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import os

from fixtures import TempDir

import testtools
from testtools.matchers import (
    DirContains,
    DirExists,
    Equals,
    FileContains,
    HasPermissions,
    )

from treeshape._treeshape import (
    DEFAULT_MODE,
    load_tree,
    )
from treeshape import (
    CONTENT,
    FileTree,
    InvalidDirectory,
    make_tree,
    PERMISSIONS,
    )


class TestMakeTree(testtools.TestCase):

    def test_directory_and_subdirectory(self):
        # If both a directory and a sub-directory are explicitly listed,
        # that's OK.
        path = self.useFixture(TempDir()).path
        make_tree(path, {'a/b/': {}, 'a/': {}})
        self.assertThat(path, DirContains(['a']))
        self.assertThat(os.path.join(path, 'a'), DirContains(['b']))
        self.assertThat(os.path.join(path, 'a', 'b'), DirExists())

    def test_empty(self):
        tempdir = self.useFixture(TempDir()).path
        make_tree(tempdir, {})
        self.assertThat(tempdir, DirContains([]))

    def test_creates_files(self):
        # When given file specifications, it creates those files underneath
        # the temporary directory.
        path = self.useFixture(TempDir()).path
        make_tree(path, {'a': {CONTENT: 'foo'}, 'b': {CONTENT: 'bar'}})
        self.assertThat(path, DirContains(['a', 'b']))
        self.assertThat(os.path.join(path, 'a'), FileContains('foo'))
        self.assertThat(os.path.join(path, 'b'), FileContains('bar'))

    def test_creates_directories(self):
        # When given directory specifications, it creates those directories.
        path = self.useFixture(TempDir()).path
        make_tree(path, {'a/': {}, 'b/': {}})
        self.assertThat(path, DirContains(['a', 'b']))
        self.assertThat(os.path.join(path, 'a'), DirExists())
        self.assertThat(os.path.join(path, 'b'), DirExists())

    def test_any_false_will_do(self):
        # If the attributes of the file is None, or an empty list or whatever,
        # treat it like an empty dict.
        path = self.useFixture(TempDir()).path
        make_tree(path, {'a/b/': None, 'c/d.txt': False})
        self.assertThat(path, DirContains(['a', 'c']))
        self.assertThat(os.path.join(path, 'a'), DirContains('b'))
        self.assertThat(os.path.join(path, 'a', 'b'), DirExists())
        self.assertThat(os.path.join(path, 'c'), DirExists())
        self.assertThat(
            os.path.join(path, 'c', 'd.txt'),
            FileContains("The file 'c/d.txt'."))

    def test_creates_parent_directories(self):
        # If the parents of a file or directory don't exist, they get created
        # too.
        path = self.useFixture(TempDir()).path
        make_tree(path, {'a/b/': {}, 'c/d.txt': {CONTENT: 'text'}})
        self.assertThat(path, DirContains(['a', 'c']))
        self.assertThat(os.path.join(path, 'a'), DirContains('b'))
        self.assertThat(os.path.join(path, 'a', 'b'), DirExists())
        self.assertThat(os.path.join(path, 'c'), DirExists())
        self.assertThat(os.path.join(path, 'c', 'd.txt'), FileContains('text'))

    def test_sets_permissions(self):
        path = self.useFixture(TempDir()).path
        make_tree(
            path, {
                'a/b/': {PERMISSIONS: 0755},
                'c/d.txt': {CONTENT: 'text', PERMISSIONS: 0600},
                })
        self.assertThat(os.path.join(path, 'a/b/'), HasPermissions('0755'))
        self.assertThat(os.path.join(path, 'c/d.txt'), HasPermissions('0600'))

    def test_directory_contents(self):
        # If CONTENT is passed to a directory, it raises InvalidDirectory.
        path = self.useFixture(TempDir()).path
        self.assertRaises(
            InvalidDirectory,
            make_tree, path, {'a/': {CONTENT: 'ignored'}})

    def test_mode(self):
        # 'mode' can be optionally passed in.  Used for implicitly created
        # directories.
        path = self.useFixture(TempDir()).path
        make_tree(
            path, {
                'a/b/': {PERMISSIONS: 0755},
                'c/d.txt': {CONTENT: 'text', PERMISSIONS: 0600}
                },
            mode=0700)
        self.assertThat(os.path.join(path, 'a/'), HasPermissions('0700'))
        self.assertThat(os.path.join(path, 'a/b/'), HasPermissions('0755'))
        self.assertThat(os.path.join(path, 'c/'), HasPermissions('0700'))


class TestLoadTree(testtools.TestCase):

    def load_tree_from(self, input_shape, mode=DEFAULT_MODE):
        path = self.useFixture(FileTree(input_shape, mode)).path
        return load_tree(path)

    def test_empty(self):
        self.assertThat(self.load_tree_from({}), Equals({}))

    def test_single_file(self):
        shape = {'foo': {CONTENT: 'content', PERMISSIONS: 0755}}
        found_shape = self.load_tree_from(shape)
        self.assertThat(found_shape, Equals(shape))

    def test_directories(self):
        shape = {'foo/': {PERMISSIONS: 0755}}
        found_shape = self.load_tree_from(shape)
        self.assertThat(found_shape, Equals(shape))

    def test_subdirectories(self):
        shape = {
            'foo/bar/': {PERMISSIONS: 0755},
            'foo/baz.txt': {CONTENT: 'content', PERMISSIONS: 0755},
            }
        found_shape = self.load_tree_from(shape, mode=0700)
        self.assertThat(
            found_shape,
            Equals(
                {'foo/': {PERMISSIONS: 0700},
                 'foo/bar/': {PERMISSIONS: 0755},
                 'foo/baz.txt': {CONTENT: 'content', PERMISSIONS: 0755},
                 }))

    def test_not_a_directory(self):
        path = self.useFixture(FileTree({'foo.txt': {}})).path
        path = os.path.join(path, 'foo.txt')
        e = self.assertRaises(ValueError, load_tree, path)
        self.assertEqual(
            "Can only load tree for directories, got %r" % (path,), str(e))



#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import os

from testtools.helpers import map_values
from testtools.matchers import (
    ContainsDict,
    Equals,
    Matcher,
    MatchesDict,
    )

from ._treeshape import (
    load_tree,
    )


def add_implicit_directories(shape):
    """
    Take a normalized shape and make sure all parent directories are included
    in it.
    """
    new_shape = dict(shape)
    # XXX: This is really slow.  We could do something better, especially
    # given that shape is sorted by name.
    for name in shape:
        for parent in iterate_parents(name):
            if parent not in new_shape:
                new_shape[parent] = {}
    return new_shape


def iterate_parents(path):
    path = path.rstrip(os.sep)
    while True:
        path = os.path.dirname(path)
        if not path:
            break
        yield path + '/'


def _is_matcher(value):
    return isinstance(value, Matcher)


def _to_matcher(value, default_matcher=Equals):
    if _is_matcher(value):
        return value
    return Equals(value)


class MatchesFileTree(Matcher):
    """Match a directory on disk against a file tree shape.

    The specified shape contains everything that is allowed to exist in the
    matched directory.  Parent directories that are implied by actual entries
    (e.g. 'foo/bar/' implies that 'foo/' exists) will be automatically added.

    Attributes can be matched using testtools matchers, e.g.::

      {'foo': {CONTENTS: Contains("Hello")}}

    Or will be simply checked for equality if they aren't matchers.  Thus::

      {'foo': {CONTENTS: Equals("Hello")}}

    is equivalent to::

      {'foo': {CONTENTS: "Hello"}}

    ``from_rough_spec`` also works with matchers, so you could call
    ``MatchesFileTree`` like this, if you want::

        MatchesFileTree(from_rough_spec([('foo', Contains("Hello")), 'bar/']))

    If the disk contains any files or directories that aren't specified in the
    matcher and aren't implied directories, then it will mismatch.  Likewise,
    if there are entries in the matcher spec that don't exist on disk, it will
    mismatch.
    """

    tree_matcher = MatchesDict

    def __init__(self, shape):
        super(MatchesFileTree, self).__init__()
        self.shape = shape

    def _matchify_attributes(self, attrs):
        """Return a matcher that matches disk attributes with expected ones.
        """
        # The pattern is that a user doesn't specify attributes they don't
        # care about.  Thus, we return a ContainsDict matcher.
        return ContainsDict(map_values(_to_matcher, attrs))

    def match(self, path):
        found_shape = load_tree(path)
        expected_shape = add_implicit_directories(self.shape)
        expected_shape = map_values(self._matchify_attributes, expected_shape)
        return self.tree_matcher(expected_shape).match(found_shape)


class HasFileTree(MatchesFileTree):
    """As ``MatchesFileTree`` but OK if some things on disk aren't specified.
    """

    tree_matcher = ContainsDict




#  treeshape: quickly make file and directory structures.
#
# Copyright (c) 2012, Jonathan Lange <jml@mumak.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

"""treeshape"""

__all__ = [
    'CONTENT',
    'DuplicateName',
    'FileTree',
    'from_rough_spec',
    'HasFileTree',
    'InvalidDirectory',
    'make_tree',
    'MatchesFileTree',
    'PERMISSIONS',
    ]


from ._dwim import (
    DuplicateName,
    from_rough_spec,
    )
from ._fixtures import FileTree
from ._matchers import (
    HasFileTree,
    MatchesFileTree,
    )
from ._treeshape import (
    CONTENT,
    InvalidDirectory,
    make_tree,
    PERMISSIONS,
    )
